@@ -272,16 +272,16 @@ G_LOCK_DEFINE_STATIC (giop_thread_list);
 static GList *giop_thread_list = NULL;
 
 static GIOPThread *
-giop_thread_new (GMainContext *context, gpointer key)
+giop_thread_new (GMainContext *context)
 {
 	GIOPThread *tdata = g_new0 (GIOPThread, 1);
 
 	tdata->lock = g_mutex_new ();
 	tdata->incoming = g_cond_new ();
 	tdata->wake_context = context;
-	tdata->key = key;
-	tdata->async_ents = g_queue_new();
-	tdata->request_queue = g_queue_new();
+	tdata->keys = NULL;
+	tdata->async_ents = NULL;
+	tdata->request_queue = NULL;
 
 	if (giop_main_thread)
 		tdata->request_handler = giop_main_thread->request_handler;
@@ -293,18 +293,53 @@ giop_thread_new (GMainContext *context, gpointer key)
 	return tdata;
 }
 
+static void
+giop_thread_key_add_T (GIOPThread *tdata, gpointer key)
+{
+	/* We don't allow a key to be reused */
+	g_assert (g_hash_table_lookup (giop_pool_hash, key) == NULL);
+  
+	tdata->keys = g_list_prepend (tdata->keys, key);
+	
+	g_hash_table_insert (giop_pool_hash, key, tdata);
+}
+
+static void
+giop_thread_key_release_T (gpointer key)
+{
+	g_hash_table_remove (giop_pool_hash, key);
+}
+
 static void
 giop_thread_free (GIOPThread *tdata)
 {
+	GList *l;
+	
 	G_LOCK (giop_thread_list);
 	giop_thread_list = g_list_remove (giop_thread_list, tdata);
 	G_UNLOCK (giop_thread_list);
 
+	if (giop_thread_safe ()) {
+		g_mutex_lock (giop_pool_hash_lock);
+		for (l = tdata->keys; l != NULL; l = l->next) {
+			giop_thread_key_release_T (l->data);
+		}
+		g_mutex_unlock (giop_pool_hash_lock);
+	}
+	
+	g_list_free (tdata->keys);
+	
 	g_mutex_free (tdata->lock);
 	g_cond_free (tdata->incoming);
-	g_queue_free (tdata->async_ents);
-	g_queue_free (tdata->request_queue);
- 
+#ifdef G_ENABLE_DEBUG
+	if (tdata->async_ents)
+		g_warning ("Leaked async ents");
+	if (tdata->request_queue)
+		g_warning ("Leaked request queue");
+#endif
+	if (tdata->invoke_policies)
+		g_queue_free (tdata->invoke_policies);
+	
 	g_free (tdata);
 }
 
@@ -319,25 +354,38 @@ giop_thread_self (void)
 		return NULL;
 
 	if (!(tdata = g_private_get (giop_tdata_private))) {
-		tdata = giop_thread_new (NULL, NULL);
+		tdata = giop_thread_new (NULL);
 		g_private_set (giop_tdata_private, tdata);
 	}
 
 	return tdata;
 }
 
-static void
-giop_thread_key_release_T (gpointer key)
+
+void
+giop_thread_key_add (GIOPThread *tdata, gpointer key)
 {
-	g_hash_table_remove (giop_pool_hash, key);
+  g_mutex_lock (giop_pool_hash_lock);
+  LINK_MUTEX_LOCK (tdata->lock);
+
+  giop_thread_key_add_T (tdata, key);
+  
+  LINK_MUTEX_UNLOCK (tdata->lock);
+  g_mutex_unlock (giop_pool_hash_lock);
 }
 
 void
 giop_thread_key_release (gpointer key)
 {
+	GIOPThread *tdata;
+	
 	if (giop_thread_safe ()) {
 		g_mutex_lock (giop_pool_hash_lock);
-		giop_thread_key_release_T (key);
+		tdata = g_hash_table_lookup (giop_pool_hash, key);
+		if (tdata != NULL) {
+			tdata->keys = g_list_remove (tdata->keys, key);
+			giop_thread_key_release_T (key);
+		}
 		g_mutex_unlock (giop_pool_hash_lock);
 	}
 }
@@ -352,19 +400,18 @@ giop_thread_request_push_key (gpointer  key,
 	g_mutex_lock (giop_pool_hash_lock);
 
 	if (!(tdata = g_hash_table_lookup (giop_pool_hash, key))) {
-		new_tdata = giop_thread_new (NULL, key);
+		new_tdata = giop_thread_new (NULL);
 		tdata = new_tdata;
+		if (key)
+			giop_thread_key_add_T (tdata, key);
 		dprintf (GIOP, "Create new thread %p for op", tdata);
 	} else
 		dprintf (GIOP, "Re-use thread %p for op", tdata);
 
 	giop_thread_request_push (tdata, poa_object, recv_buffer);
 
-	if (new_tdata) {
-		if (key)
-			g_hash_table_insert (giop_pool_hash, key, new_tdata);
+	if (new_tdata) 
 		g_thread_pool_push (giop_thread_pool, tdata, NULL);
-	}
 
 	g_mutex_unlock (giop_pool_hash_lock);
 }
@@ -393,9 +440,18 @@ giop_mainloop_handle_input (GIOChannel     *source,
 			    gpointer        data)
 {
 	char c;
+	GIOPThread *tdata = giop_thread_self ();
 
 	read (WAKEUP_POLL, &c, sizeof (c));
-	giop_recv_handle_queued_input ();
+
+	LINK_MUTEX_LOCK (tdata->lock);
+	while (!giop_thread_queue_empty_T (tdata)) {
+		LINK_MUTEX_UNLOCK (tdata->lock);
+		giop_thread_queue_process (tdata);
+		LINK_MUTEX_LOCK (tdata->lock);
+	}
+	LINK_MUTEX_UNLOCK (tdata->lock);
+
 
 	return TRUE;
 }
@@ -404,6 +460,7 @@ static void
 giop_request_handler_thread (gpointer data, gpointer user_data)
 {
 	gboolean done;
+	GList *l;
 	GIOPThread *tdata = data;
 
 	g_private_set (giop_tdata_private, tdata);
@@ -411,14 +468,17 @@ giop_request_handler_thread (gpointer data, gpointer user_data)
 	dprintf (GIOP, "Thread %p woken to handle request", tdata);
 
 	do {
-		giop_thread_request_process (tdata);
+		giop_thread_queue_process (tdata);
 
 		g_mutex_lock (giop_pool_hash_lock);
 		LINK_MUTEX_LOCK (tdata->lock);
 
-		if ((done = g_queue_is_empty (tdata->request_queue)) &&
-		    tdata->key)
-			giop_thread_key_release_T (tdata->key);
+		if ((done = giop_thread_queue_empty_T (tdata))) {
+			for (l = tdata->keys; l != NULL; l = l->next)
+				giop_thread_key_release_T (l->data);
+			g_list_free (tdata->keys);
+			tdata->keys = NULL;
+		}
 
 		LINK_MUTEX_UNLOCK (tdata->lock);
 		g_mutex_unlock (giop_pool_hash_lock);
@@ -446,11 +506,12 @@ giop_init (gboolean threaded, gboolean blank_wire_data)
 	if (threaded) {
 		GIOPThread *tdata;
 
-		/* FIXME: should really cleanup with descructor */
-		giop_tdata_private = g_private_new (NULL);
+		/* We need a destructor to clean up if giopthreads are used
+		 * outside of ORBit controlled threads */
+		giop_tdata_private = g_private_new ((GDestroyNotify)giop_thread_free);
 
 		giop_main_thread = tdata = giop_thread_new (
-			g_main_context_default (), NULL); /* main thread */
+			g_main_context_default ()); /* main thread */
 
 		if (pipe (corba_wakeup_fds) < 0) /* cf. g_main_context_init_pipe */
 			g_error ("Can't create CORBA main-thread wakeup pipe");
@@ -481,7 +542,11 @@ wakeup_mainloop (void)
 	char c = 'A'; /* magic */
 	int  res;
 	while ((res = write (WAKEUP_WRITE, &c, sizeof (c))) < 0  &&
-	       (errno == EAGAIN || errno == EINTR));
+	       errno == EINTR );
+#ifdef G_ENABLE_DEBUG
+	if (res < 0 && errno == EAGAIN)
+		g_warning ("Looks like you have no glib mainloop running");
+#endif
 	if (res < 0)
 		g_error ("Failed to write to GIOP wakeup socket %d 0x%x(%d) (%d)",
 			 res, errno, errno, WAKEUP_WRITE);
@@ -519,28 +584,19 @@ giop_invoke_async (GIOPMessageQueueEntry *ent)
 	if (!giop_thread_io ())
 		ent->async_cb (ent);
 
-	else if (ent->src_thread == giop_main_thread) {
-
-		if (giop_thread_self () == giop_main_thread)
-			ent->async_cb (ent);
-
-		else {
-			GIOPThread *tdata = ent->src_thread;
-
-			g_mutex_lock (tdata->lock); /* ent_lock */
-
-			buf = NULL;
-			g_queue_push_tail (tdata->async_ents, ent);
-			
-			g_assert (tdata->wake_context);
-			giop_incoming_signal_T (tdata);
-
-			g_mutex_unlock (tdata->lock); /* ent_unlock */
-		}
-	} else { /* Push callback to a new thread */
-		g_warning ("FIXME: emit async callback in it's own thread ?");
-
+	else if (ent->src_thread == giop_thread_self ())
 		ent->async_cb (ent);
+
+	else {
+		GIOPThread *tdata = ent->src_thread;
+		
+		g_mutex_lock (tdata->lock); /* ent_lock */
+
+		buf = NULL;
+		tdata->async_ents = g_list_prepend (tdata->async_ents, ent);
+		giop_incoming_signal_T (tdata);
+		
+		g_mutex_unlock (tdata->lock); /* ent_unlock */
 	}
 
 	/* NB. At the tail end of async_cb 'Ent' is invalid / freed */
@@ -594,17 +650,123 @@ typedef struct {
 	gpointer recv_buffer;
 } GIOPQueueEntry;
 
-void
-giop_thread_request_process (GIOPThread *tdata)
+/* this sucks, we need a wider scale re-factor */
+#include "../orb-core/orbit-policy.h"
+#include "orbit/poa/poa-types.h"
+
+static GList *
+first_valid_request (GIOPThread *tdata, gboolean *no_policy)
 {
+	GList *l;
+	ORBitPolicy *policy;
+
+	if (!tdata->invoke_policies || !tdata->invoke_policies->head) {
+		*no_policy = TRUE;
+		return NULL;
+	}
+
+	*no_policy = FALSE;
+
+	for (l = tdata->request_queue; l; l = l->next) {
+		int i;
+		GIOPQueueEntry *qe = l->data;
+		ORBit_POAObject pobj = qe->poa_object;
+
+		for (i = 0; i < policy->allowed_poas->len; i++)
+			if (g_ptr_array_index (policy->allowed_poas, i) == pobj->poa)
+				return l;
+	}
+
+	return NULL;
+}
+
+gboolean
+giop_thread_queue_empty_T (GIOPThread *tdata)
+{
+	gboolean no_policy;
+	
+	if (first_valid_request (tdata, &no_policy))
+		return FALSE;
+
+	else if (no_policy)
+		return (!tdata->request_queue &&
+			!tdata->async_ents);
+
+	else
+		return TRUE;
+}
+
+static gpointer
+giop_list_pop (GList **list)
+{
+	gpointer p;
+
+	if (!*list)
+		return NULL;
+	
+	p = (*list)->data;
+	*list = g_list_delete_link (*list, *list);
+
+	return p;
+}
+
+void
+giop_thread_queue_process (GIOPThread *tdata)
+{
+	GIOPMessageQueueEntry *ent;
 	GIOPQueueEntry *qe = NULL;
+	GList   *request;
+	gboolean no_policy;
 
-	LINK_MUTEX_LOCK (tdata->lock);
-	qe = g_queue_pop_head (tdata->request_queue);	
-	LINK_MUTEX_UNLOCK (tdata->lock);
+	if (!tdata)
+		tdata = giop_thread_self ();
 
-	if (qe)
+	request = first_valid_request (tdata, &no_policy);
+
+	dprintf (MESSAGES, "handle queued input [%p], (%d)\n", request, no_policy);
+
+	LINK_MUTEX_LOCK (tdata->lock); /* ent_lock */
+
+	if (no_policy)
+		ent = giop_list_pop (&tdata->async_ents);
+	else
+		ent = NULL;
+
+	if (!ent) {
+		if (no_policy)
+			qe = giop_list_pop (&tdata->request_queue);
+
+		else if (request) {
+			qe = request->data;
+			tdata->request_queue = g_list_delete_link (tdata->request_queue, request);
+		}
+	}
+
+	dprintf (MESSAGES, "Queue pop %p, %p, %d", ent, qe);
+	
+	LINK_MUTEX_UNLOCK (tdata->lock); /* ent_unlock */
+
+	if (ent)
+		giop_invoke_async (ent);
+
+	if (qe) {
 		tdata->request_handler (qe->poa_object, qe->recv_buffer, NULL);
+		g_free (qe);
+	}
+}
+
+void
+giop_thread_queue_tail_wakeup (GIOPThread *tdata)
+{
+	if (!tdata)
+		return; /* FIXME: no I/O thread */
+
+	LINK_MUTEX_LOCK (tdata->lock); /* ent_lock */
+
+	if ((tdata->request_queue || tdata->async_ents) && tdata->wake_context)
+		wakeup_mainloop ();
+
+	LINK_MUTEX_UNLOCK (tdata->lock); /* ent_unlock */
 }
 
 void
@@ -627,7 +789,7 @@ giop_thread_request_push (GIOPThread *tdata,
 
 	LINK_MUTEX_LOCK (tdata->lock);
 
-	g_queue_push_tail (tdata->request_queue, qe);
+	tdata->request_queue = g_list_append (tdata->request_queue, qe);
 	giop_incoming_signal_T (tdata);
 
 	LINK_MUTEX_UNLOCK (tdata->lock);
