@@ -23,15 +23,12 @@ static GMutex *object_lock = NULL;
 static guint
 g_CORBA_Object_hash (gconstpointer key)
 {
-	guint        retval;
 	CORBA_Object obj = (gpointer) key;
 
-	retval = obj->type_qid;
-
+	/* type_id is not reliable: cf. corbaloc */
 	g_assert (obj->object_key != NULL);
-	retval ^= IOP_ObjectKey_hash (obj->object_key);
 
-	return retval;
+        return IOP_ObjectKey_hash (obj->object_key);
 }
 
 static gboolean
@@ -247,17 +244,15 @@ ORBit_objref_get_proxy (CORBA_Object obj)
 }
 
 static gboolean
-ORBit_try_connection (CORBA_Object obj)
+ORBit_try_connection_T (CORBA_Object obj)
 {
 	gboolean retval = FALSE;
-	gboolean try = FALSE;
 	LinkConnectionStatus status;
+	LinkConnection *cnx = LINK_CONNECTION (obj->connection);
 
-	if (!obj->connection)
-		return FALSE;
+	OBJECT_UNLOCK (obj);
 
-	status = link_connection_wait_connected
-		((LinkConnection *) obj->connection);
+	status = link_connection_wait_connected (cnx);
 
 	switch (status) {
 	case LINK_CONNECTING:
@@ -268,12 +263,17 @@ ORBit_try_connection (CORBA_Object obj)
 		break;
 	case LINK_DISCONNECTED:
 		/* Have a go at reviving it */
-		dprintf (MESSAGES, "re-connecting dropped cnx %p", obj->connection);
-		giop_connection_try_reconnect (obj->connection);
+		dprintf (MESSAGES, "re-connecting dropped cnx %p", cnx);
+		if (giop_connection_try_reconnect (GIOP_CONNECTION (cnx)) == LINK_CONNECTED)
+			retval = TRUE;
 		break;
 	}
 
-	return TRUE;
+	OBJECT_LOCK (obj);
+
+	g_assert (LINK_CONNECTION (obj->connection) == cnx);
+
+	return retval;
 }
 
 GIOPConnection *
@@ -289,11 +289,17 @@ ORBit_object_get_connection (CORBA_Object obj)
 	GIOPConnection *cnx = NULL;
 
 	OBJECT_LOCK (obj);
-	if (ORBit_try_connection (obj)) {
-		cnx = obj->connection;
-		giop_connection_ref (cnx);
-		OBJECT_UNLOCK (obj);
-		return cnx;
+
+	if (obj->connection) {
+		if (ORBit_try_connection_T (obj)) {
+			cnx = obj->connection;
+			giop_connection_ref (cnx);
+			OBJECT_UNLOCK (obj);
+			return cnx;
+		} else {
+			OBJECT_UNLOCK (obj);
+			return NULL;
+		}
 	}
   
 	g_assert (obj->connection == NULL);
@@ -316,7 +322,7 @@ ORBit_object_get_connection (CORBA_Object obj)
 				obj->orb, proto, host, service,
 				is_ssl ? LINK_CONNECTION_SSL : 0, iiop_version);
 
-			if (ORBit_try_connection (obj)) {
+			if (obj->connection && ORBit_try_connection_T (obj)) {
 				obj->object_key = objkey;
 				obj->connection->orb_data = obj->orb;
 
@@ -550,44 +556,69 @@ ORBit_demarshal_object (CORBA_Object   *obj,
 	return FALSE;
 }
 
-CORBA_boolean
-CORBA_Object_is_a (CORBA_Object       obj,
-		   const CORBA_char  *logical_type_id,
-		   CORBA_Environment *ev)
+CORBA_char*
+ORBit_object_to_corbaloc (CORBA_Object       obj,
+                          CORBA_Environment *ev)
 {
-	static GQuark  corba_object_quark = 0;
-	static GQuark  omg_corba_object_quark = 0;
-	CORBA_boolean  retval;
-	gpointer       args[] = { (gpointer *)&logical_type_id };
-	GQuark         logical_type_quark;
+        CORBA_char      *retval = NULL;
 
-	if (!corba_object_quark)
-		corba_object_quark = g_quark_from_static_string (
-			"IDL:CORBA/Object:1.0");
+        if (!obj) {
+                dprintf (OBJECTS, "Corbaloc NIL object\n");
+                return CORBA_string_dup ("corbaloc::/");
+        }
 
-	if (!omg_corba_object_quark)
-		omg_corba_object_quark = g_quark_from_static_string (
-			"IDL:omg.org/CORBA/Object:1.0");
+        OBJECT_LOCK (obj);
 
-	logical_type_quark = g_quark_from_string (logical_type_id);
+        if (!obj->profile_list) {
+                IOP_generate_profiles (obj);
+                ORBit_register_objref (obj);
+        }
 
-	if (logical_type_quark == corba_object_quark)
-		return CORBA_TRUE;
+        if (!(retval = ORBit_corbaloc_from (obj->profile_list,
+                                            obj->object_key))) {
+                CORBA_exception_set_system (
+                        ev, ex_CORBA_BAD_PARAM, CORBA_COMPLETED_NO);
+                /* FIXME, set minor code with vendor specific id */
+        }
+ 
+        OBJECT_UNLOCK (obj);
+         
+        return retval;
+}
 
-	if (logical_type_quark == omg_corba_object_quark)
-		return CORBA_TRUE;
 
-	if (!obj)
-		return CORBA_FALSE;
+CORBA_Object
+ORBit_object_by_corbaloc  (CORBA_ORB          orb,
+                           const gchar       *corbaloc,
+                           CORBA_Environment *ev)
+{
+        CORBA_Object  retval       = CORBA_OBJECT_NIL;
+        GSList       *profile_list = NULL;
 
-	if (logical_type_quark == obj->type_qid)
-		return CORBA_TRUE;
+        g_return_val_if_fail (orb!=NULL,      CORBA_OBJECT_NIL);
+        g_return_val_if_fail (corbaloc!=NULL, CORBA_OBJECT_NIL);
+        g_return_val_if_fail (ev!=NULL,       CORBA_OBJECT_NIL);
 
-	ORBit_small_invoke_stub (
-		obj, &CORBA_Object__imethods[4],
-		&retval, args, NULL, ev);
+	if (!strncmp (corbaloc, "corbaloc::/", 1 + strlen ("corbaloc::/")))
+		return CORBA_OBJECT_NIL;  
 
-	return retval;
+        if (!(profile_list = ORBit_corbaloc_parse (corbaloc)))  {
+                CORBA_exception_set_system (
+                        ev, ex_CORBA_BAD_PARAM, CORBA_COMPLETED_NO);
+                /* FIXME, set minor code with vendor specific id */
+                return CORBA_OBJECT_NIL;
+        }
+
+        if (!(retval = ORBit_objref_find (orb, "", profile_list))) {
+                CORBA_exception_set_system (
+                        ev, ex_CORBA_BAD_PARAM, CORBA_COMPLETED_NO);
+                /* FIXME, set minor code with vendor specific id */
+
+                IOP_delete_profiles (orb, &profile_list);
+                return CORBA_OBJECT_NIL;
+        }
+
+        return retval;
 }
 
 static gboolean
@@ -617,6 +648,50 @@ ORBit_impl_CORBA_Object_is_a(PortableServer_ServantBase *servant,
 	*(CORBA_boolean *)ret = ORBit_IInterface_is_a (ci->idata, type_id);
 }
 
+CORBA_boolean
+CORBA_Object_is_a (CORBA_Object       obj,
+		   const CORBA_char  *logical_type_id,
+		   CORBA_Environment *ev)
+{
+	static GQuark  corba_object_quark = 0;
+	static GQuark  omg_corba_object_quark = 0;
+	CORBA_boolean  retval;
+	gpointer       servant;
+	gpointer       args[] = { (gpointer *)&logical_type_id };
+	GQuark         logical_type_quark;
+
+	if (!corba_object_quark)
+		corba_object_quark = g_quark_from_static_string (
+			"IDL:CORBA/Object:1.0");
+
+	if (!omg_corba_object_quark)
+		omg_corba_object_quark = g_quark_from_static_string (
+			"IDL:omg.org/CORBA/Object:1.0");
+
+	logical_type_quark = g_quark_from_string (logical_type_id);
+
+	if (logical_type_quark == corba_object_quark)
+		return CORBA_TRUE;
+
+	if (logical_type_quark == omg_corba_object_quark)
+		return CORBA_TRUE;
+
+	if (!obj)
+		return CORBA_FALSE;
+
+	if (logical_type_quark == obj->type_qid)
+		return CORBA_TRUE;
+
+	if ((servant = ORBit_small_get_servant (obj)))
+		ORBit_impl_CORBA_Object_is_a (servant, &retval, args, NULL, ev, NULL);
+
+	else /* warning: obeys POA policies */
+		ORBit_small_invoke_stub (obj, &CORBA_Object__imethods[4],
+					 &retval, args, NULL, ev);
+
+	return retval;
+}
+
 static void
 ORBit_impl_ORBit_get_type_id (PortableServer_ServantBase *servant,
 			      gpointer ret, gpointer *args,
